(*  :Title:	Differentiable Functions Measuring Filter Characteristics *)

(*  :Authors:	Brian L. Evans *)

(*
    :Summary:   This file derives differentiable functions describing
		the magnitude response, phase response, quality,
		and percent overshoot step response of analog filters.
 *)

(*  :Context:	FilterOptimization`Filters`  *)

(*  :PackageVersion:  FOP 1.0	@(#)Filters.m	1.4	11/20/97  *)

(*
   :Copyright:

Copyright (c) 1993-1997 The Regents of the University of California.
All rights reserved.

Permission is hereby granted, without written agreement and without
license or royalty fees, to use, copy, modify, and distribute this
software and its documentation for any purpose, provided that the above
copyright notice and the following two paragraphs appear in all copies
of this software.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY 
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES 
ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF 
THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF 
SUCH DAMAGE.

THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS.
 *)

(*  :History:	*)

(*  :Keywords:	*)

(*  :Source:	*)

(*  :Warning:	*)

(*  :Mathematica Version:  2.0  *)

(*  :Limitation:  *)

(*
    :Discussion: 

	We actually employ the arctangent function for two different reasons:
	  
	  (1) to compute the argument of the frequency response
	      to give us the phase response, and
	  (2) to solve the equation a cos(c t) + b sin(c t) = 0.

	The issue is that there are two flavors of the arctangent function,
	  Arctan[z] and Arctan[x, y].  For (1), we define our own arg
	  function.  For (2), we simply use ArcTan[z].
 *)

(*  :Functions:	*)


BeginPackage[ "FilterOptimization`Filters`" ]


(*  U S A G E     I N F O R M A T I O N  *)

linearPhaseSlope::usage =
"linearPhaseSlope[r, i, n, wp] computes the best slope of the line
that approximates the phase response of an analog filter.
The analog filter has n complex pole pairs, r[1] + I i[1],
r[1] - I i[1], r[2] + I i[2], r[2] - I i[2], ...";

magnitude::usage =
"magnitude[r, i, n, w] returns a differentiable function describing
the magnitude response of an analog filter as a function of
frequency w. 
The analog filter has n complex pole pairs, r[1] + I i[1],
r[1] - I i[1], r[2] + I i[2], r[2] - I i[2], ...";

peakTimeApprox::usage =
"peakTimeApprox[r, i, n] approximates the time at which the step
response of an analog filter hits its maximum value.
For a second-order section, this value is exact.
The analog filter has n complex pole pairs, r[1] + I i[1],
r[1] - I i[1], r[2] + I i[2], r[2] - I i[2], ...";

percentOvershoot::usage =
"percentOvershoot[r, i, n] indicates the overall percent overshoot
for the analog filter.
The analog filter has n complex pole pairs, r[1] + I i[1],
r[1] - I i[1], r[2] + I i[2], r[2] - I i[2], ...";

percentOvershootInSection::usage =
"percentOvershootInSection[rp, ip, A, B] returns a differentiable
function describing the percent overshoot of the step response of
a second-order analog filter section with transfer function
(A s + B) / (s^2 - 2 rp s + rp^2 + ip^2 ), i.e., the poles
are located at -rp +/- ip.";

phase::usage =
"phase[r, i, n, w] returns a differentiable function describing
the phase response of an analog filter as a function of
frequency w.
The analog filter has n complex pole pairs, r[1] + I i[1],
r[1] - I i[1], r[2] + I i[2], r[2] - I i[2], ...";

quality::usage =
"quality[r, i, n] returns a differentiable function describing
the quality factor of the analog filter.
The analog filter has n complex pole pairs, r[1] + I i[1],
r[1] - I i[1], r[2] + I i[2], r[2] - I i[2], ...
The quality factor for each pole pair ranges from 0.5 to infinity.
At 0.5, the pole pair is on the real axis.
At infinity, the pole pair is on the imaginary axis.
For robust filters, the quality factors should generally be less
than 20 and the geometric mean of the quality factors should be
less than 4.";

qualityOneSection::usage =
"qualityOneSection[r, i, j] returns the formula for the quality
of one second-order section with its complex poles located at
 r[j] + I i[j] and r[j] - I i[j].";

stepResponse::usage =
"stepResponse[r, i, n, t] returns the step response of the analog
filter having n complex pole pairs, r[1] + I i[1], r[1] - I i[1],
r[2] + I i[2], r[2] - I i[2], ...";

(*  E N D     U S A G E     I N F O R M A T I O N  *)


Begin["`Private`"]

(* decide how we are going to find the argument of a x + I y  *)
argfun[ z_ ] := ArcTan[z]
argfun[ x_, y_ ] := ArcTan[y/x]

(* computes filter magnitude as a differentiable function of w, the frequency
   -- For each pole pair, the mag resp. is 
      ( r[j]^2 + i[j]^2 ) /
      ( Sqrt[r[j]^2 + (w - i[j])^2] Sqrt[r[j]^2 + (w + i[j])^2] )
   -- I have rewritten this function to have better numerical properties
      by expand the denominator, combining terms under the square root sign,
      and putting the polynomial in Horner's form:
	( r[j]^2 + i[j]^2 ) /
	Sqrt[ w^2 ( w^2 + 2 (r[j]^2 - i[j]^2) ) + ( i[j]^2 + r[j]^2 )^2 ]
 *)
magnitude[ r_, i_, n_, w_ ] :=
	Module[	{denom, denomSquared, j, numer, s},
		numer = Product[ ( r[j]^2 + i[j]^2 ), {j, 1, n} ];
		denomSquared = 
			Product[ w^2 ( w^2 + 2 (r[j]^2 - i[j]^2) ) +
				 ( i[j]^2 + r[j]^2 )^2,
			 	 {j, 1, n} ];
		numer / Sqrt[ denomSquared ] ]

(* computes filter phase as a differentiable function of w, the frequency *)
phase[ r_, i_, n_, w_ ] :=
	Module[	{j},
		Sum[ argfun[ r[j], w - i[j] ] + argfun[ r[j], w + i[j] ],
		     {j, 1, n} ] ]

(* computes the geometric mean of the quality factors *)
(* -- remember that r[j] is negative, so we need -2 r[j] because Q >= 0.5 *)
qualityOneSection[ r_, i_, j_ ] :=
	Sqrt[r[j]^2 + i[j]^2] / (-2 r[j])
quality[ r_, i_, n_ ] :=
	Module[ {j}, Product[ qualityOneSection[r, i, j], {j, 1, n} ]^(1/n) ]

(* percentOvershoot *)
(* -- computes partial fractions decomposition for each second-order section:

                    c s + d
        ------------------------------
        (s - rp + I ip)(s - rp - I ip)

      by computing the part. frac. coefficient for a one-pole section:

              A                 Aconj
      ---------------  +  ----------------  ==
      (s - rp - I ip)     ( s - rp + I ip)

        I A ip - I Aconj ip - A rp - Aconj rp + A s + Aconj s
	-----------------------------------------------------
	       (ip + I rp - I s) (ip - I rp + I s)

      so c = A + Aconj = 2 Amag cos(Aphase) and
         d = I ip (A - Aconj) - rp (A + Aconj)
           = -2 ip Amag sin(Aphase) - 2 rp Amag cos(Aphase).
      There are no constraints on c or d other than d cannot equal 0.

   -- A is computed by partial fractions decomposition of the transfer function
   -- Partial fractions decomposition returns 1/s ( F1(s) + F2(s)+ ... Fn(s) )
      where Fi(s) is the transfer function of a second-order section.
   -- This function returns the worst case percentOvershoot as the
      (sum of overshoot for each second-order section) + n - 1.
 *)
percentOvershoot[ r_, i_, n_ ] :=
	Module[ {Amag, Aphase, c, d, j},
		Sum[ Amag = partFracCoeffAbs[r, i, n, j];
		     Aphase = partFracCoeffPhase[r, i, n, j];
		     c = 2 Amag Cos[Aphase];
		     d = -2 Amag ( i[j] Sin[Aphase] + r[j] Cos[Aphase] );
		     percentOvershootInSection[ r[j], i[j], c, d ],
		     {j, 1, n} ] ]

(* percentOvershootInSection *)
(* -- assumes a general second-order section with two poles and one zero:

                    A s + B                    B     (A wn^2 / B) s + wn^2
        ------------------------------   =   ----  --------------------------
        (s - rp + I ip)(s - rp - I ip)       wn^2  s^2 - 2 rp s + rp^2 + ip^2

	with wn^2 = rp^2 + ip^2 which is the natural frequency of pole pair
	where rp is real part of pole and ip is imag part of pole.

   -- overshoot calculations have been worked out for the normalized section
   -- To find maximum value of the step response, we take its derivative w/r
      to time, set it to zero, and solve for tpeak.  Solution for tpeak is
      periodic with period Pi because it solves  c1 cos(b t) + c2 sin(b t) = 0
      which requires an inverse tangent function.  The symbolic answer (from
      Maple) for tpeak is off by one period:
 
    > solve( a*cos(ip*t) + (rp/ip)*(rp+a)*sin(ip*t) + ip*sin(ip*t) = 0, t );

      -(1/ip) arctan( a ip / ( rp^2 + a rp + ip^2 ) )

      where a = (A wn^2 / B).  This can be verified by setting B = wn^2 and
      A = 0 (so a = 0) as well as wd = -ip and sigma = -rp, and then by
      comparing the answer to [Franklin, p. 59].

	  G. F. Franklin, J. D. Powell, and A. Emami-Naeini, {Feedback Control
	of Dynamic Systems}, Addison-Wesley, 1986.

	The time value for the peak of the time response is:

	   tpeak = ( ArcTan[rp^2 + a rp + ip^2, a ip] + Pi ) / (-ip);

      where a = (A wn^2 / B).  Substituting for a, the expression simplifies to

	   tpeak = ( ArcTan[ B + A ip, A ip ] + Pi ) / (- ip);
 *)
percentOvershootInSection[ rp_, ip_, A_, B_ ] :=
	Module[ {a, stepResp, t, tpeak, wnSquared},
		wnSquared = rp^2 + ip^2;
		a = A wnSquared / B;
		stepResp = 1 - Exp[rp t] ( Cos[ip t] - (rp + a)/ip Sin[ip t] );
		tpeak = ( ArcTan[ A ip / ( B + A rp ) ] + Pi ) / (- ip);
		100 ( ( stepResp /. t -> tpeak ) - 1 ) ]

(*
percentOvershootInSection[ rp_, ip_, A_, B_ ] :=
	Module[ {a, tpeak},
		a = A (rp^2 + ip^2) / B;
		tpeak = ( ArcTan[ A ip / ( B + A rp ) ] + Pi ) / (- ip);
		100 Exp[rp tpeak] Sqrt[ ((rp + a)^2 + ip^2) / (rp^2 + ip^2) ] ]
 *)

(* peakTimeApprox -- takes average of peak times of each second-order section *)
peakTimeApprox[ r_, i_, n_ ] :=
	Module[ {Aphase, j},
		Sum[ Aphase = partFracCoeffPhase[r, i, n, j];
		     ( Aphase + Pi/2 ) / (- i[j]),
		     {j, 1, n} ] / n ]

stepResponse2ndOrderSection[ rp_, ip_, A_, B_, t_ ] :=
	Module[ {gamma, resp, wnSquared},
		wnSquared = rp^2 + ip^2;
		gamma = A wnSquared / B;
		resp = 1 - Exp[rp t] ( Cos[ip t] - (rp + gamma)/ip Sin[ip t] );
		resp B / wnSquared ]

stepResponse[ r_, i_, n_, t_ ] :=
	Module[ {Amag, Aphase, c, d, j},
		Sum[ Amag = partFracCoeffAbs[r, i, n, j];
		     Aphase = partFracCoeffPhase[r, i, n, j];
		     c = 2 Amag Cos[Aphase];
		     d = -2 Amag ( i[j] Sin[Aphase] + r[j] Cos[Aphase] );
		     stepResponse2ndOrderSection[ r[j], i[j], c, d, t ],
		     {j, 1, n} ] ]

(* returns mag of partial fraction coefficient for one pole in pole pair k  *)
(* -- by convention, r[j] and i[j] are negative so we have the -2 i[j] term *)
(* -- must be a module in case k is symbolic and equal to j *)
partFracCoeffAbs[ r_, i_, n_, k_ ] :=
	Module[ {arg, denomsquared, j, numer},
		numer = Product[ ( r[j]^2 + i[j]^2 ), {j, 1, n} ] / (-2 i[k]);
		arg = ((r[k] - r[j])^2 + (i[k] - i[j])^2) *
		      ((r[k] - r[j])^2 + (i[k] + i[j])^2);
		denomsquared =
		      Product[ arg, {j, 1, k-1} ] Product[ arg, {j, k+1, n} ];
		numer / Sqrt[ denomsquared ] ]

(* returns phase of partial fraction coefficient for one pole in pole pair k *)
(* -- phase for k = j is phase of 1/(2 i[k] I), with i[k] < 0 *)
(* -- must be a module in case k is symbolic and equal to j *)
partFracCoeffPhase[ r_, i_, n_, k_ ] :=
	Module[ {arg, j},
		arg = - argfun[ (r[k] - r[j]), (i[k] - i[j]) ] -
			argfun[ (r[k] - r[j]), (i[k] + i[j]) ];
		Pi/2 + Sum[ arg, {j, 1, k-1} ] + Sum[ arg, {j, k+1, n} ] ]

(*  phaseIntegral  *)
(*  -- integral of w phase( H(jw) ) from w = 0 .. wmax for one pole pair *)
phaseIntegral[ r1_, i1_, wmax_ ] :=
-( 2*r1*wmax +
   (i1^2 - r1^2 - wmax^2)*( argfun[r1, wmax - i1] + argfun[r1, wmax + i1] ) +
   i1*r1*Log[1 + (wmax - i1)^2/r1^2] - i1*r1*Log[1 + (wmax + i1)^2/r1^2] ) / 2

(*  linearPhaseSlope  *)
(*  -- computes integral of w phase( H(jw) ) from w = 0 .. wmax divided	*)
(*     by (1/3) wmax^3 = the integral of w^2 from w = 0 .. wmax		*)
linearPhaseSlope[ r_, i_, n_, wmax_ ] :=
	Module[	{int, j, slope, total},
		int = phaseIntegral[ r[1], i[1], wmax ];
		total = Sum[ int /. { r[1]->r[j], i[1]->i[j] }, {j, 1, n} ];
		total / ((1/3)*wmax^3) ]

(*  deriveLinearPhaseSlope  *)
(*  -- derives the linearPhaseSlope formula automatically for different n *)
(*  -- there is a manual derivation in the file "derivations/manual.m"    *)
deriveLinearPhaseSlope[ r_, i_, n_, wp_] :=
	Module[ {w},
		Integrate[ w phase[r, i, n, w], {w, 0, wp} ] /
		  Integrate[ w^2, {w, 0, wp} ] ]


End[]
EndPackage[]

Protect[ linearPhaseSlope,		magnitude,
	 peakTimeApprox, 		percentOvershoot,
	 percentOvershootInSection,	phase,
	 quality,			qualityOneSection,
	 stepResponse ];

Null
